unit CACOVARS; {Unit 8 of CONCRETE.PAS}

interface

  uses
    GLOBALS, MATHS, SCREEN, UTILS, POST, LEGENDS, Graph, Dos, Crt;

  function PartialNumerator (var theta: shortVector;
                  var kappa: complex;
                  var q: integer): complex;

  procedure PartialDenoms (var kappa: complexVector;
                  var denom: complex;
                  p, j: integer);

  procedure PartialCoefficients (var partialC: complexVector;
                  var LSDESpec: ARMARecord);

  procedure GammaCoeffs (var gCoeff: complexVector;
                  var LSDESpec: ARMARecord);

  function SecondOrderCAVC (var LSDESpec: ARMARecord;
                  var tau: real): real;

  procedure RealcCovar (var alpha, beta: real;
                  var kappa: complexVector;
                  var gammaC: shortVector;
                  lags: integer);

  procedure CACVOrdinates (var LSDESpec: ARMARecord;
                  lags: integer;
                  var acvBOX: rBOX;
                  var SCREENBOX: IBOX;
                  var cACVector: longVector);

  procedure PlotContinuousAcovars (var SCREENBOX, POSTBOX: IBOX;
                  var LSDESpec: ARMARecord;
                  var lags: integer);

  procedure SampleCcovars (var LSDESpec: ARMARecord;
                  var lags: integer);

implementation

{_________________________________________________}

  function PartialNumerator (var theta: shortVector;
                  var kappa: complex;
                  var q: integer): complex;

{The numertor part of the coefficient of the partial fraction decomposition of}
{the rational transfer function of a Linear Stochastic Dufferential Equation via}
{Horner's method of nested multiplication: (theta0*k + theta1)*k + theta2)*k . . . }

    var
      polyVal: complex;
      i: integer;

  begin
    polyVal.re := theta[0];
    polyVal.im := 0.0;
    for i := 1 to q do
      begin
        ComplexMult(polyVal, kappa, polyVal);
        polyVal.re := polyVal.re + theta[i];
      end;
    PartialNumerator := polyVal;
  end; {PartialNumerator}

{_________________________________________________}

  procedure PartialDenoms (var kappa: complexVector;
                  var denom: complex;
                  p, j: integer);

{The denominator part of the jth coefficient ot the partial}
{fraction decomposition of the rational transfer function}
{of a Linear Stochastic Differential Equation.}

    var
      i: integer;
      factor: complex;

  begin
    denom.re := 1.0;
    denom.im := 0.0;
    for i := 1 to p do
      begin
        if i <> j then
          begin
            ComplexSub(kappa[j], kappa[i], factor);
            ComplexMult(denom, factor, denom);
          end;
      end;
  end; {PartialDenoms}

{_________________________________________________}

  procedure PartialCoefficients (var partialC: complexVector;
                  var LSDESpec: ARMARecord);

{The coefficients of the partial fraction decomposition of the}
{rational transfer function of a Linear Stochastic Differential}
{Equation}

    var
      theta: shortVector;
      kappa: complexVector;
      denominator, numerator: complex;
      p, q, j: integer;

  begin
    p := LSDESpec.p;
    q := LSDESpec.q;
    kappa := LSDESpec.poles;
    theta := LSDESpec.beta;

    for j := 1 to p do
      begin
        PartialDenoms(kappa, denominator, p, j);
        numerator := PartialNumerator(theta, kappa[j], q);
        ComplexDiv(numerator, denominator, partialC[j]);
      end;
  end; {PartialCoefficients;}

{_________________________________________________}

  procedure GammaCoeffs (var gCoeff: complexVector;
                  var LSDESpec: ARMARecord);

    var
      i, j, p: integer;
      numerator, denominator, ratio: complex;
      kappa, partialC: complexVector;

{The coefficients of the exponential terms of}
{the continuous autocovariance  function}

  begin
    p := LSDESpec.p;
    kappa := LSDESpec.poles;

    PartialCoefficients(partialC, LSDESpec);

    for i := 1 to p do
      begin {i}
        gCoeff[i].re := 0.0;
        gCoeff[i].im := 0.0;
        for j := 1 to p do
          begin {j}
            ComplexMult(partialC[i], partialC[j], numerator);
            ComplexAdd(kappa[i], kappa[j], denominator);
            ComplexDiv(numerator, denominator, ratio);
            ComplexAdd(gCoeff[i], ratio, gCoeff[i]);
          end; {j}
      end; {i}
  end; {GammaCoeffs}

{_________________________________________________}

  function SecondOrderCAVC (var LSDESpec: ARMARecord;
                  var tau: real): real;

{This is the prototpe of the general procedure for finding the continuous time ACV}
{which employs complex arithmatic}

    var
      numerator, denominator, omega, gamma, autoCov: real;
      prod, sum: real;

  begin
    if LSDESpec.pr = 1 then
      begin
        omega := LSDESpec.poles[1].im;
        gamma := LSDESpec.poles[1].re;
        autoCov := sin(omega * tau) / (2 * omega);
        autoCov := autoCov - cos(omega * tau) / (2 * gamma);
        autoCov := autoCov * Exp(gamma * tau);
        sum := omega * omega + gamma * gamma;
        autoCov := autoCov / sum;
      end
    else if LSDESpec.p = 2 then
      begin
        prod := 2 * LSDESpec.poles[1].re * LSDESpec.poles[2].re;
        sum := Sqr(LSDESpec.poles[1].re) - sqr(LSDESpec.poles[2].re);
        denominator := prod * sum;
        numerator := LSDESpec.poles[2].re * Exp(LSDESpec.poles[1].re * tau);
        numerator := numerator - LSDESpec.poles[1].re * Exp(LSDESpec.poles[2].re * tau);
        autoCov := numerator / denominator;
      end
    else if LSDESpec.p = 1 then
      begin
        numerator := Exp(LSDESpec.poles[1].re * tau);
        denominator := 2 * LSDESpec.poles[1].re;
        autoCov := -numerator / denominator;
      end;

    SecondOrderCAVC := autoCov;
  end; {SecondOrderCAVC}

{_________________________________________________}

  procedure RealcCovar (var alpha, beta: real;
                  var kappa: complexVector;
                  var gammaC: shortVector;
                  lags: integer);

{This procedure finds the autocovarinces of a CARMA(2, 1) process}
{with complex conjugate poles. It avoids complex arithmatic.}

    var
      omega, delta, expdtau: real;
      a2Pb2, a2Mb2, d2Pw2, twoabd, twoabw: real;
      tau, sinMult, cosMult, extraCos: real;
      i: integer;

  begin
    omega := kappa[1].im;
    delta := kappa[1].re;

    for i := 0 to lags do
      begin
        tau := i; {Type conversion}
        a2Pb2 := alpha * alpha + beta * beta;
        a2Mb2 := alpha * alpha - beta * beta;
        d2Pw2 := delta * delta + omega * omega;
        twoabd := 2.0 * alpha * beta * delta;
        twoabw := 2.0 * alpha * beta * omega;

        sinMult := twoabd - omega * a2Mb2;
        sinMult := sinMult * Sin(omega * tau);

        cosMult := twoabw + delta * a2Mb2;
        cosMult := cosMult * Cos(omega * tau);

        extraCos := a2Pb2 / delta;
        extraCos := extraCos * cos(omega * tau);

        expdtau := Exp(delta * tau);
        gammaC[i] := (sinMult - cosMult) / d2Pw2;
        gammaC[i] := (gammaC[i] - extraCos) * expdtau;
      end; {i}
  end; {RealcCovar}

{_________________________________________________}

  procedure CACVOrdinates (var LSDESpec: ARMARecord;
                  lags: integer;
                  var acvBOX: rBOX;
                  var SCREENBOX: IBOX;
                  var cACVector: longVector);
    var
      i, j, p, BOXWDT: integer;
      tau, gamma, omega, amplitude, increment: real;
      gCoeff: complexVector;
      trigTerm, autoComplex: complex;

      {CACVOrdinates calculates the vector of ordinates for}
      {plotting the continuous LSDE autocorrelation function}

  begin

{We need to sample the continuous autocovariance function at integer laGS}

    BOXWDT := SCREENBOX.X2 - SCREENBOX.X1;

    acvBOX.xMax := lags; {the maximum value of x in the data vector}
    acvBOX.xMin := 0.0; {the minmum value  of x in the data vector}
    acvBOX.yMax := 0.0; {the maximum value of y in the Data vector}
    acvBOX.yMin := 0.0; {the minimum value of y in the Data vector}
    tau := 0.0;
    increment := lags / BOXWDT;

    if LSDESpec.p < 2 then
      for i := 0 to BOXWDT do
        begin {i}
          cACVector[i] := SecondOrderCAVC(LSDESpec, tau);
          tau := tau + increment;
          if cACVector[i] > acvBOX.yMax then
            acvBOX.yMax := cACVector[i];
          if cACVector[i] < acvBOX.yMin then
            acvBOX.yMin := cACVector[i];
        end; {i}

{Note that we are by-passing the code for the second-order case}

    if LSDESpec.p >= 2 then
      begin {p >= 2}
        p := LSDESpec.p;
        GammaCoeffs(gCoeff, LSDESpec);
        tau := 0.0;
        increment := lags / BOXWDT;
        for i := 0 to BOXWDT do
          begin {i}
            autoComplex.re := 0.0;
            autoComplex.im := 0.0;
            for j := 1 to p do
              begin {j}
                gamma := LSDESpec.poles[j].re;
                omega := LSDESpec.poles[j].im;
                amplitude := Exp(gamma * tau);
                trigTerm.re := -amplitude * cos(omega * tau);
                trigTerm.im := -amplitude * sin(omega * tau);
                ComplexMult(gCoeff[j], trigTerm, trigTerm);
                ComplexAdd(autoComplex, trigTerm, autoComplex);
              end; {j}
            cACVector[i] := autoComplex.re;
            if i > 0 then
              cACVector[i] := cACVector[i]/cACVector[0];
            if cACVector[i] < acvBOX.yMin then
              acvBOX.yMin := cACVector[i];
            tau := tau + increment;
          end; {i}
        cACVector[0] := 1.0;
        acvBOX.yMax :=  1.0;
      end; {p >= 2}

  end; {CACVOrdinates}

{_________________________________________________}

  procedure PlotContinuousAcovars (var SCREENBOX, POSTBOX: IBOX;
                  var LSDESpec: ARMARecord;
                  var lags: integer);

    var
      acvBOX: rBOX;
      cacVector: longvector;
      postFileName, legend: string;
      F: text;

  begin
    CACVOrdinates(LSDESpec, lags, acvBOX, SCREENBOX, cacVector);

    GetGraphics(graphicsOn, SCREENBOX, currentMode);
    acvBOX.xMin := 0.0;
    acvBOX.xMaX := lags;
    ScreenFrame(acvBOX, SCREENBOX);
    ScreenPlotMod(acvBOX, SCREENBOX, cacVector);
    legend := LSDEAcovarLegend(LSDESpec);
    MultiCaption(legend);

    CallForReturn;
    Readln;
    CloseGraph;
    RestoreCrtMode;
    ClrScr;

    if PSPict then
      begin
        Write('    Name the PostScript File = ');
        Readln(postFileName);
        if postFileName <> '' then
          begin {postFileName}
            Assign(F, postFileName);
            Rewrite(F);
            PostHeader(F, POSTBOX, TeXFormat); {no postExtras}
            PostFrame(F, acvBOX, POSTBOX);
            PostPlotMod(F, acvBOX, POSTBOX, cacVector);
            PostFooter(F, POSTBOX, legend, TexFormat);
            Close(F);
          end; {postFileName}
      end; {PSPict}

  end; {PlotContinuousAcovars}
{_________________________________________________}

  procedure SampleCcovars (var LSDESpec: ARMARecord;
                  var lags: integer);

{This provides the continuous autocovariance sampled at the integer points.}
{Uses GammaCoeffs(gCoeff, LSDESpec)}

    var
      tau, j, p: integer;
      gamma, omega, amplitude: real;
      gCoeff: complexVector;
      trigTerm, autoComplex: complex;

  begin

    p := LSDESpec.p;
    GammaCoeffs(gCoeff, LSDESpec);

    for tau := 0 to lags do
      begin {tau}
        autoComplex.re := 0.0;
        autoComplex.im := 0.0;
        for j := 1 to p do
          begin {j}
            gamma := LSDESpec.poles[j].re;
            omega := LSDESpec.poles[j].im;
            amplitude := Exp(gamma * tau);
            trigTerm.re := -amplitude * cos(omega * tau);
            trigTerm.im := -amplitude * sin(omega * tau);
            ComplexMult(gCoeff[j], trigTerm, trigTerm);
            ComplexAdd(autoComplex, trigTerm, autoComplex);
          end; {j}
        LSDESpec.aCovar[tau] := autoComplex.re;
      end; {tau}

    LSDESpec.aCovarsFound := true;

  end; {SampleCcovars}

{________________________________________________}

end. {CAC0VARS: : Unit 8 of CONCRETE.PAS}

