unit CONVERTS; {Unit 12 of CONCRETE.PAS}

interface

  uses
    GLOBALS, MATHS, UTILS, LEGENDS, ACQUIRE, CACOVARS, DACOVARS, OPTIMUM, Dos, Crt, Graph;

  procedure YuleWalker (p, q: integer;
                  gamma: shortVector;
                  var alpha: shortVector;
                  var varEpsilon: real);

  procedure Minit (var mu: shortVector;
                  var varEpsilon: real;
                  gamma: shortVector;
                  q: integer);

  procedure ARMAparameters (p, q: integer;
                  gamma: shortvector;
                  var alpha, mu: shortVector;
                  var varEpsilon: real);

  procedure ComplexPower (var cartRoot: complex;
                  var cPower: complex;
                  var pwr: integer);

  procedure CpartialFractDecomp (var cBeta, cMu, cD: complexVector;
                  var p, q: integer);

  procedure ARMAtoCARMA (var ARMASpec, cARMASpec: ARMARecord;
                  var cNumerator, cDenominator: complexVector);

  procedure LSDEtoEDLM (var LSDESpec, EDLMSpec: ARMARecord);

  procedure ARMAtoLSDE (var ARMASpec, LSDESpec: ARMARecord);

implementation

{_________________________________________________}

  procedure YuleWalker (p, q: integer;
                  gamma: shortVector;
                  var alpha: shortVector;
                  var varEpsilon: real);

    var
      a: matrix;
      b: shortVector;
      i, j, k: integer;

  begin {Yule-Walker}
    for i := 1 to p do
      begin {i}
        b[i] := -gamma[q + i];
        for j := 1 to p do
          begin {j}
            k := abs(q + i - j);
            a[i, j] := gamma[k];
          end{j}
      end; {i}

    LUsolve(1, p, a, alpha, b);
    alpha[0] := 1;
    varEpsilon := 0;
    for i := 0 to p do
      varEpsilon := varEpsilon + alpha[i] * gamma[i];

  end; {Yule-Walker}

{_________________________________________________}

  procedure Minit (var mu: shortVector;
                  var varEpsilon: real;
                  gamma: shortVector;
                  q: integer);

    var
      D: matrix;
      delta, f: shortVector;
      i, j, iterations: integer;
      convergence: boolean;


{-------------------------------------------}

    function CheckDelta (q: integer;
                    var delta, mu: shortVector): boolean;

      var
        i: integer;
        muNorm, deltaNorm: real;

    begin
      muNorm := 0.0;
      deltaNorm := 0.0;
      for i := 0 to q do
        begin
          muNorm := muNorm + sqr(mu[i]);
          deltaNorm := deltaNorm + Sqr(delta[i])
        end;
      if (deltaNorm / muNorm) > 0.00001 then
        CheckDelta := false
      else
        CheckDelta := true;

    end; {Checkdelta}

{------------------------------------------}

  begin {PROCEDURE Minit}

{initialise the vector mu }
    for i := 0 to q do
      mu[i] := 0.0;
    mu[0] := Sqrt(gamma[0]);
    convergence := false;
    iterations := 0;

    while (convergence = false) and (iterations < 10) do
      begin

{Form the matrix of derivatives}
        for i := 0 to q do
          for j := 0 to q do
            begin
              D[i, j] := 0.0;
              if (j - i) >= 0 then
                D[i, j] := mu[j - i];
              if (i + j) <= q then
                D[i, j] := D[i, j] + mu[i + j];
            end;

{Find the function value}
        for j := 0 to q do
          begin
            f[j] := gamma[j];
            for i := 0 to q - j do
              f[j] := f[j] - mu[i] * mu[i + j]
          end;

{Find the updating vector}
        LUsolve(0, q + 1, D, delta, f);

{Update the value of mu}
        for i := 0 to q do
          mu[i] := mu[i] + delta[i];
        iterations := iterations + 1;

{Check for convergence}
        convergence := CheckDelta(q, delta, mu);
      end; {while}
    Writeln;
    Writeln('    The number of iterations was ', iterations : 3);
    Writeln;

{Renomalise the results}
    varEpsilon := 1.0;
    for i := 1 to q do
      begin {i}
        mu[i] := mu[i] / mu[0];
        varEpsilon := varEpsilon + mu[i] * mu[i];
      end; {i}
    mu[0] := 1;
    varEpsilon := gamma[0] / varEpsilon;

  end; {Minit}

{_________________________________________________}

  procedure ARMAparameters (p, q: integer;
                  gamma: shortVector;
                  var alpha, mu: shortVector;
                  var varEpsilon: real);

{This pocedure infers the ARMA parmeters from the autocovariance function.}
{It uses the iterative procure Minit in finding the moving average parameters.}
{In case it is used in finding the EDLM from the LSDE, the AR parameters are}
{already available.}

    var
      t, i, j, k: integer;
      psi: shortVector;

  begin {ARMAparameters}

    for t := 0 to q do
      begin {t}
        psi[t] := 0.0;
        for i := 0 to p do
          for j := 0 to p do
            begin {i, j}
              k := abs(t + j - i);
              psi[t] := psi[t] + alpha[i] * alpha[j] * gamma[k];
            end; {i, j}
      end;{t}

    if q > 0 then
      Minit(mu, varEpsilon, psi, q)

  end;{ARMAparameters}

{_________________________________________________}

  procedure ComplexPower (var cartRoot: complex;
                  var cPower: complex;
                  var pwr: integer);
    var
      argu, modu, moduPower, arguTimes: real;
      polarRoot: polar;

  begin
    ComplexToPolar(cartRoot, polarRoot);

    modu := polarRoot.modu;
    argu := polarRoot.argu;
    moduPower := Power(modu, pwr);
    arguTimes := Pwr * argu;
    cPower.re := moduPower * Cos(arguTimes);
    cPower.im := moduPower * Sin(arguTimes);
  end; {ComplexPower}

{_________________________________________________}

  procedure CpartialFractDecomp (var cBeta, cMu, cD: complexVector;
                  var p, q: integer);

{This procedure finds the coefficients of a partial fraction decomposition}
{of a rational function given the coefficients of the numerator polynomial}
{in the vector beta and the complex-valued inverse roots of the denominator}
{polynomial in the vector mu}

    var
      numerator, denominator, factor, store: complex;
      i, j, r: integer;

  begin
    for j := 1 to p do
      begin {j}

{calculate the jth numerator by Horner's method}
        r := p - 1 - q;
        ComplexPower(cMu[j], factor, r);
        numerator.re := cBeta[0].re;
        numerator.im := cBeta[0].im;

        for i := 1 to q do
          begin
            ComplexMult(numerator, cMu[j], numerator);
            ComplexAdd(numerator, cBeta[i], numerator);
{numerator := numerator * mu[j] + beta[i];}
          end;
        ComplexMult(numerator, factor, numerator);
{numerator := numerator * factor;}

{calculate the jth denominator}
        denominator.re := 1.0;
        denominator.im := 0.0;
        for i := 1 to p do
          begin {i}
            if (i <> j) then
              begin {i<>j}
                ComplexSub(cMu[j], cMu[i], store);
                ComplexMult(denominator, store, denominator);
{denominator := denominator * (mu[j] - mu[i]);}
              end; {i<>j}
          end; {i}
{d[j] := numerator / denominator;}
        ComplexDiv(numerator, denominator, cD[j]);
      end; {j}

  end; {CpartialFractDecomp}

{_________________________________________________}

  procedure ComplexNumerator (var cKappa, cD, cNumerator: complexVector;
                  p: integer);

{This procedure composes the numerator of a rational tranfer}
{function from its partial fractions with an allowance for the}
{possibility that the poles and zeros are complex numbers.}

    var
      i, j, k, q: integer;
      store: complex;
      coeffs: complexVector;

  begin
    for i := 0 to p - 1 do
      begin {i}
        cNumerator[i].re := 0.0;
        cNumerator[i].im := 0.0;
      end; {i}

    for j := 1 to p do
      begin {j}
        coeffs[0].re := 1.0;
        coeffs[0].im := 0.0;
        q := 1;
        for k := 1 to p do
          begin {k}
            if (k <> j) then
              begin
                coeffs[q].re := 0.0;
                coeffs[q].im := 0.0;
                for i := q downto 1 do
                  begin {i}
                    ComplexMult(cKappa[k], coeffs[i - 1], store);
                    ComplexSub(coeffs[i], store, coeffs[i]);
                  end; {i}
                q := q + 1;
              end; {j<>k}
          end; {k}

        for i := 0 to p - 1 do
          begin {i}
            ComplexMult(cD[j], coeffs[i], store);
            ComplexAdd(cNumerator[i], store, cNumerator[i]);
          end; {i}
      end; {j}

  end; {ComplexNumerator}

{_________________________________________________}

  procedure ARMAtoCARMA (var ARMASpec, CARMASpec: ARMARecord;
                  var cNumerator, cDenominator: complexVector);

    var
      p, q, i: integer;
      cBeta, CD: complexVector;
      degToRad: real;

{Conversion of ARMA to CARMA via the principle of Impulse Invariance}

  begin
    degToRad := ArcTan(1.0) / 45.0;

    p := ARMASpec.p;
    q := ARMASpec.q;
    CARMASpec.p := p;
    CARMASpec.q := p - 1;

    for i := 0 to q do
      begin
        cBeta[i].re := ARMASpec.beta[i];
        cBeta[i].im := 0.0;
      end;

{Find the partial fraction decomposition of the ARMA operator}
{cD are the partial fraction numerators}
    with ARMASpec do
      CpartialFractDecomp(cBeta, ARMASpec.poles, cD, p, q);

{Convert to discrete time poles to poles of a differential equation}
    for i := 1 to p do
      begin
        CARMASpec.polarPoles[i].argu := ARMASpec.polarPoles[i].argu * degToRad;
        CARMASpec.polarPoles[i].modu := ARMASpec.polarPoles[i].modu;
        ComplexLog(CARMASpec.polarPoles[i], CARMASpec.poles[i]);
      end;
    CARMASpec.rootsFound := true;

    ComplexNumerator(CARMASpec.poles, cD, cNumerator, p);
    InverseRootsToCoeffs(p, CARMASpec.poles, cDenominator);

    for i := 0 to p do
      cARMASpec.alpha[i] := cDenominator[i].re;

    for i := 0 to q do
      cARMASpec.beta[i] := cNumerator[i].re;

    CARMASpec.noModel := false;
    CARMASpec.ordersSpecified := true;
    CARMASpec.parametersFound := true;
    CARMASpec.rootsFound := true;

  end;{ARMAtoCARMA}

{_________________________________________________}

  procedure LSDEtoEDLM (var LSDESpec, EDLMSpec: ARMARecord);

    var
      i, p, q, lags: integer;
      radToDeg: real;
      dDenominator: complexVector;

  begin
    radToDeg := 45.0 / Arctan(1.0);
    LSDESpec.q := LSDESpec.p - 1;
    p := LSDESpec.p;
    q := p - 1;
    lags := p + q;

{0. Obtain sample values from the LSDE autocovariance function}
    if LSDESpec.aCovarsFound = false then
      SampleCcovars(LSDESpec, lags);
    EDLMSpec.aCovar := LSDESpec.aCovar;

{1. Derive the discrete-time poles from the continuous-time poles}
    for i := 1 to p do
      begin {i}
        ComplexAntilog(LSDESpec.poles[i], EDLMSpec.poles[i]);
        ComplexToPolar(EDLMSpec.poles[i], EDLMSpec.polarPoles[i]);
        EDLMSpec.polarPoles[i].argu := EDLMSpec.polarPoles[i].argu * radToDeg;
      end; {i}

{2. Find the discrete-time denominator}
    InverseRootsToCoeffs(p, EDLMSpec.poles, dDenominator);
    for i := 0 to p do
      EDLMSpec.alpha[i] := dDenominator[i].re;

    EDLMSpec.p := p;
    EDLMSpec.q := q;
{3. Find the discrete-time MA parameters via the Cramer-Wold decomposition}
    with EDLMSpec do
      ARMAparameters(p, q, aCovar, alpha, beta, varEpsilon);

    EDLMSpec.noModel := false;
    EDLMSpec.ordersSpecified := true;
    EDLMSpec.parametersFound := true;
  end; {LSDEtoEDLM}

{_________________________________________________}

  procedure ARMAtoLSDE (var ARMASpec, LSDESpec: ARMARecord);

    var
      lags, i, nOfTrials, theMenuItem: integer;
      gamma: shortVector;
      complexAlpha: complexVector;
      trialFunct, degToRad: real;
      trialPt: sPoint;

  begin
    degToRad := ArcTan(1.0) / 45;
    LSDESpec.p := ARMASpec.p;
    LSDESpec.q := LSDESpec.p-1;
    lags := ARMASpec.p + ARMASpec.q;
    if (ARMASpec.aCovarsFound = false) then
      ARMACovariances(ARMASpec, gamma, lags);

{Convert from discrete-Stime poles to poles of a differential equation}

    if (LSDESpec.rootsFound = false) then {OR SHOULD IT BE POLES FOUND ?}
      begin {if}
        for i := 1 to ARMASpec.p do
          begin {i}
            LSDESpec.polarPoles[i].argu := ARMASpec.polarPoles[i].argu * degToRad;
            LSDESpec.polarPoles[i].modu := ARMASpec.polarPoles[i].modu;
            ComplexLog(LSDESpec.polarPoles[i], LSDESpec.poles[i]);
          end; {i}
        LSDESpec.p := ARMASpec.p;
        LSDESpec.rootsFound := true;
        InverseRootsToCoeffs(LSDESpec.p, LSDESpec.poles, complexAlpha);
      end; {if}

    for i := 0 to ARMASpec.p do
      LSDESpec.alpha[i] := complexAlpha[i].re;

{We ought to use ARMASpec.Acovars instead of gamma  within Funct in OPTIMUM}

    repeat
      ClrScr;

      Writeln;
      Writeln('    STARTING VALUES FOR THE MINIMISATION');
      Writeln;
      Writeln('    The Nelder--Mead algorithm finds the LSDE parameters that match');
      Writeln('    match the LSDE autocovariances to the ARMA autocovariances. The');
      Writeln('    algorithm requires some starting values for alpha and beta.');
      Writeln;
      Writeln('      1. Use the default starting values');
      Writeln;
      Writeln('      2. Specify the starting values');
      Writeln;
      Writeln('                      Select a number and press RETURN');

      GetBoundedInteger(1, 2, theMenuItem);
    until theMenuItem in [1, 2];

    case theMenuItem of
      1: 
        begin
          trialFunct := 1000.0;
          trialPt[1] := 1.0;
          trialPt[2] := 0.0;
        end;
      2: 
        begin
          trialFunct := 1000.0;
          Writeln('      Specify the starting value for alpha');
          ReadANumber(trialPt[1]);
          Writeln('      Specify the starting value for beta');
          ReadANumber(trialPt[1]);
        end;
    end;{cases}

{Use the Nelder--Mead algorithm  to find the LSDE parameters that}
{match the sampled LSDE autocovariances to the ARMA autocovariances.}

    Writeln;
    Amoeba(trialFunct, trialPt, ARMASpec, LSDESpec, lags, nOfTrials);

    ClrScr;

    Writeln;
    Writeln('    RESULTS OF THE OPTIMISATION PROCEDURE');
    Writeln;
    Writeln('    The number of function evaluations = ', nOfTrials : 3);
    Writeln;
    Writeln('    The minimum value of the criteion function = ', trialFunct : 4 : 6);
    Writeln;
    Writeln('    The parameters of the LSDE(2, 1) autocovariance function');
    Writeln;
    Writeln('     a = ', trialPt[1] : 4 : 6);
    Writeln('     b = ', trialPt[2] : 4 : 6);

    READLN;

    with LSDESpec do
      begin
        beta[0] := 2 * trialPt[1];
        beta[1] := 2 * (trialPt[1] * poles[1].re + trialPt[2] * poles[1].im);
        noModel := false;
      end;

      LSDESpec.noModel := false;

  end; {ARMAtoLSDE}

{_________________________________________________}

end. {CONVERTS: Unit 12 of CONCRETE.PAS}

