unit OPTIMUM; {Unit 11 of CONCRETE.PAS}

interface

  uses
    GLOBALS, MATHS, UTILS, CACOVARS, DACOVARS;

  function Funct (var trialPt: sPoint;
                  var ARMASpec, LSDESpec: ARMARecord;
                  var lags, trials: integer): real;

  procedure InitialiseSimplex (var simplex: sMatrix;
                  var fvalue: sPoint;
                  var trialPt: sPoint;
                  var step: real;
                  var lags, trials: integer);

  procedure FindCentroid (var simplex: sMatrix;
                  var centroid: sPoint;
                  var worstIndex, lags: integer);

  procedure FindNewPt (scalar: real;
                  var firstPt, secondPt, newPt: sPoint;
                  var lags: integer);

  procedure FindExtremePts (var fvalue: sPoint;
                  var worstIndex, poorIndex, bestIndex, lags: integer);

  function sConvergence (var fvalue: sPoint;
                  var worstIndex, bestIndex: integer;
                  epsilon: real): boolean;

  procedure Amoeba (var trialFunct: real;
                  var trialPt: sPoint;
                  var ARMASpec, LSDESpec: ARMARecord;
                  var lags, trials: integer);

implementation

{_________________________________________________}

  function Funct (var trialPt: sPoint;
                  var ARMASpec, LSDESpec: ARMARecord;
                  var lags, trials: integer): real;

{This criterion function calculates the sum of squares of the differnces between}
{the ARMA and LSDE autocovariances  at the integer  points 0, 1, . . . ,  p + q,}
{corresponding to the LSDE parameters plus the variance of the forcing function.}
{Here, lags := ARMASpec.p + ARMASpec.q.}

{The parameter trialPt: sPoint conveys the parameters alpha, beta to}
{the function <RealcCovar>, which generates the LSDE autocovariances.}

    var
      kappa: complexVector;
      gammaC, gammaD: shortVector;
      alpha, beta, measure: real;
      i: integer;


  begin


{lags := ARMASpec.p + ARMASpec.q;}
{if (ARMASpec.aCovarsFound = false) then}
{ARMACovariances(ARMASpec, gamma, lags)}
{else}

    for i := 0 to lags do
      gammaD[i] := ARMASpec.ACovar[i];

    for i := 1 to ARMASpec.p do
      kappa[i] := LSDESpec.poles[i];
    alpha := trialPt[1];
    beta := trialPt[2];
    trials := trials + 1;

    RealcCovar(alpha, beta, kappa, gammaC, lags);
    measure := 0.0;
    for i := 0 to lags do
      measure := measure + Sqr(gammaC[i] - gammaD[i]);
    Funct := measure;

  end; {Funct}

{_________________________________________________}

  procedure InitialiseSimplex (var simplex: sMatrix;
                  var fvalue: sPoint;
                  var trialPt: sPoint;
                  var step: real;
                  var lags, trials: integer);
    var
      i: integer;

{The initial simplex is constructed by taking lags deviations from}
{the base sPoint along the orthogonal Cartesian axes. to create its}
{remaining lags vertices. A function evaluation is performed at each}
{of the vertices.}

  begin
    simplex[0] := trialPt;
    fvalue[0] := Funct(simplex[0], ARMASpec, LSDESpec, lags, trials);
    for i := 1 to lags do
      begin
        simplex[i] := simplex[0];
        simplex[i, i] := simplex[i, i] + step;
        fvalue[i] := Funct(simplex[i], ARMASpec, LSDESpec, lags, trials)
      end;
  end; {InitialiseSimplex}

{_________________________________________________}

  procedure FindCentroid (var simplex: sMatrix;
                  var centroid: sPoint;
                  var worstIndex, lags: integer);

{worstIndex is the index of the worst vertex sPoint of the simplex.}
{That is the sPoint with the highest function value.}

    var
      i, j: integer;

  begin
    for j := 1 to lags do
      begin {j: the index of the vertex}
        centroid[j] := 0.0;
        for i := 0 to lags do
          begin
            if (i <> worstIndex) then
              centroid[j] := centroid[j] + simplex[i, j];
          end; {i}
        centroid[j] := centroid[j] / lags;
      end; {j}
  end; {FindCentroid}

{_________________________________________________}

  procedure FindNewPt (scalar: real;
                  var firstPt, secondPt, newPt: sPoint;
                  var lags: integer);
    var
      j: integer;

{This procedure is used for finding the trial points: the reflection sPoint,}
{the extension sPoint, the contraction points, and shrinkage sPoint.}

  begin
    for j := 1 to lags do
      newPt[j] := (1.0 + scalar) * firstPt[j] - scalar * secondPt[j];

  end; {FindNewPt}

{_________________________________________________}

  procedure FindExtremePts (var fvalue: sPoint;
                  var worstIndex, poorIndex, bestIndex, lags: integer);

{worstIndex identifies the worst vertex, poorIndex identifies}
{the next worst vertex and bestIndex identifies the best.}

    var
      i: integer;

  begin
    worstIndex := 0;
    poorIndex := 0;
    bestIndex := 0;

    for i := 1 to lags do
      begin {i}
        if fvalue[i] < fvalue[bestIndex] then
          bestIndex := i;
        if fvalue[i] > fvalue[worstIndex] then
          worstIndex := i;
      end; {i}

    poorIndex := bestindex;
    for i := 0 to lags do
      begin {i}
        if (i <> worstIndex) then
          begin {if}
            if fvalue[i] > fvalue[poorIndex] then
              poorIndex := i
          end; {if}
      end; {i}

  end; {FindExtremePts}

{_________________________________________________}

  function sConvergence (var fvalue: sPoint;
                  var worstIndex, bestIndex: integer;
                  epsilon: real): boolean;
    var
      testValu, fbest, fworst: real;

  begin
    fworst := fvalue[worstIndex];
    fbest := fvalue[bestIndex];
    testValu := 2.0 * Abs(fworst - fbest);
    testValu := testValu / (Abs(fworst) + Abs(fbest));
    sConvergence := (testValu < epsilon);
  end; {sConvergence}

{_________________________________________________}

  procedure Amoeba (var trialFunct: real;
                  var trialPt: sPoint;
                  var ARMASpec, LSDESpec: ARMARecord;
                  var lags, trials: integer);

    const
      alpha = 1.0;
      beta = 0.5;
      gamma = 2.0;
      epsilon = 1.0e-8;
      maxTrials = 500;

    var
      simplex: sMatrix;
      stopIt, shrinkage: boolean;
    var
      fvalue, extraPt, centroid: sPoint;
      extraFunct, step: real;
      i, worstIndex, poorIndex, bestIndex: integer;

  begin
    step := 0.5;
    stopIt := false;
    trials := 0;

    InitialiseSimplex(simplex, fvalue, trialPt, step, lags, trials);

    repeat {until sConvergence}
      shrinkage := false;
      FindExtremePts(fvalue, worstIndex, poorIndex, bestIndex, lags);
      FindCentroid(simplex, centroid, worstIndex, lags);
      FindNewPt(alpha, centroid, simplex[worstIndex], trialPt, lags);
      trialFunct := Funct(trialPt, ARMASpec, LSDESpec, lags, trials);

      if trialFunct < fvalue[bestIndex] then {(1)}
        begin {Reflection and Extension}
          FindNewPt(gamma, centroid, simplex[worstIndex], extraPt, lags);
          extraFunct := Funct(extraPt, ARMASpec, LSDESpec, lags, trials);
          if extraFunct < trialFunct then
            begin {Extension}
              simplex[worstIndex] := extraPt;
              fvalue[worstIndex] := extraFunct;
            end {Extension}
          else
            begin {Reflection}
              simplex[worstIndex] := trialPt;
              fvalue[worstIndex] := trialFunct;
            end; {Reflection}
        end {Reflection and Extension}

      else if trialFunct < fvalue[poorIndex] then {(2)}
        begin
          simplex[worstIndex] := trialPt;
          fvalue[worstIndex] := trialFunct;
        end

      else if (fvalue[poorIndex] < trialFunct) and (trialFunct < fvalue[worstIndex]) then  {(3)}
        begin {Outside Contraction}
          FindNewPt(beta, centroid, simplex[worstIndex], trialPt, lags);
          trialFunct := Funct(trialPt, ARMASpec, LSDESpec, lags, trials);
          if trialFunct < fvalue[worstIndex] then
            begin {Contraction}
              simplex[worstIndex] := trialPt;
              fvalue[worstIndex] := trialFunct;
            end {Contraction}
          else
            Shrinkage := true
        end {Outside Contraction}

      else if (trialFunct > fvalue[worstIndex]) then {(4)}
        begin {Inside Contraction}
          FindNewPt(-beta, centroid, simplex[worstIndex], trialPt, lags);
          trialFunct := Funct(trialPt, ARMASpec, LSDESpec, lags, trials);
          if trialFunct < fvalue[worstIndex] then
            begin {Contraction}
              simplex[worstIndex] := trialPt;
              fvalue[worstIndex] := trialFunct;
            end {Contraction}

          else
            Shrinkage := true
        end; {Indside Contraction}

      if Shrinkage then
        for i := 0 to lags do
          begin {Shrinkage}
            FindNewPt(-beta, simplex[i], simplex[bestIndex], simplex[i], lags);
            fvalue[i] := Funct(simplex[i], ARMASpec, LSDESpec, lags, trials);
          end; {Shrinkage}

      stopIt := sConvergence(fvalue, worstIndex, bestIndex, epsilon);
      if trials > maxTrials then
        stopIt := true;

    until stopIt;
    trialPt := simplex[bestIndex];
    trialFunct := fvalue[bestIndex];

  end; {Ameoba}

{_________________________________________________}


end. {OPTIMIUM:  Unit 11 of CONCRETE.PAS}

